import sys
import os

try:
    from ansible.executor import playbook_executor
    from ansible.inventory import Inventory
    from ansible.parsing.dataloader import DataLoader
    from ansible.vars import VariableManager
    from ansible.utils.display import Display
    display = Display()
except ImportError:
    logger.error("Ansible is not installed")
    sys.exit(1)


class Options(object):
    def __init__(self, **kwargs):
        props = (
            'ask_pass', 'ask_sudo_pass', 'ask_su_pass', 'ask_vault_pass',
            'become_ask_pass', 'become_method', 'become', 'become_user',
            'check', 'connection', 'diff', 'extra_vars', 'flush_cache',
            'force_handlers', 'forks', 'inventory', 'listhosts', 'listtags',
            'listtasks', 'module_path', 'module_paths',
            'new_vault_password_file', 'one_line', 'output_file',
            'poll_interval', 'private_key_file', 'python_interpreter',
            'remote_user', 'scp_extra_args', 'seconds', 'sftp_extra_args',
            'skip_tags', 'ssh_common_args', 'ssh_extra_args', 'subset', 'sudo',
            'sudo_user', 'syntax', 'tags', 'timeout', 'tree',
            'vault_password_files', 'verbosity')

        for p in props:
            if p in kwargs:
                setattr(self, p, kwargs[p])
            else:
                setattr(self, p, None)

class Runner(object):
    def __init__(self, playbook, hosts='hosts', options={}, passwords={}, vault_pass=None):

        # Set options
        self.options = Options()
        for k, v in options.iteritems():
            setattr(self.options, k, v)

        # Set global verbosity
        self.display = display
        self.display.verbosity = self.options.verbosity

        # Executor has its own verbosity setting
        playbook_executor.verbosity = self.options.verbosity

        # Gets data from YAML/JSON files
        self.loader = DataLoader()
        # Set vault password
        if vault_pass is not None:
            self.loader.set_vault_password(vault_pass)
        elif 'VAULT_PASS' in os.environ:
            self.loader.set_vault_password(os.environ['VAULT_PASS'])

        # All the variables from all the various places
        self.variable_manager = VariableManager()
        if self.options.python_interpreter is not None:
            self.variable_manager.extra_vars = {
                'ansible_python_interpreter': self.options.python_interpreter
            }

        # Set inventory, using most of above objects
        self.inventory = Inventory( loader=self.loader, 
                                    variable_manager=self.variable_manager,
                                    host_list=hosts)

        if len(self.inventory.list_hosts()) == 0:
            # Empty inventory
            self.display.error("Provided hosts list is empty.")
            sys.exit(1)

        self.inventory.subset(self.options.subset)

        if len(self.inventory.list_hosts()) == 0:
            # Invalid limit
            self.display.error("Specified limit does not match any hosts.")
            sys.exit(1)

        self.variable_manager.set_inventory(self.inventory)

        # Setup playbook executor, but don't run until run() called
        self.pbex = playbook_executor.PlaybookExecutor(
            playbooks=[playbook],
            inventory=self.inventory,
            variable_manager=self.variable_manager,
            loader=self.loader,
            options=self.options,
            passwords=passwords)

    def run(self):
        # Run Playbook and get stats
        self.pbex.run()
        stats = self.pbex._tqm._stats

        return stats



from yaml.representer import Representer
import collections
import logging
import yaml
import sys
import os

from bridgy import inventory
from bridgy.config.base import ConfigBase

logger = logging.getLogger()

CONFIG_TEMPLATE_NAME = 'sample_config_2.yml'


class Config(ConfigBase):

    def sources(self):
        sources = self.dig('inventory', 'source')
        ret = []
        for srcCfg in sources:
            source = srcCfg['type']
            ret.append( (source, srcCfg, ) )
        return ret

    @property
    def version(self):
        return 2

    @property
    def config_template_path(self):
        return CONFIG_TEMPLATE_NAME

    def verify(self):
        super(Config, self).verify()

        for source, srcCfg in self.sources():
            if source == 'aws':
                if 'name' not in srcCfg:
                    logger.error("AWS inventory sources must specify name.")
                    sys.exit(1)
                if 'profile' in srcCfg:
                    if 'access_key_id' in srcCfg or 'secret_access_key' in srcCfg or 'session_token' in srcCfg:
                        logger.error("AWS profile conflicts with access_key_id, secret_access_key, or session_token.")
                        sys.exit(1)
                if 'access_key_id' in srcCfg or 'secret_access_key' in srcCfg or 'session_token' in srcCfg:
                    if 'profile' in srcCfg:
                        logger.error("AWS access_key_id, secret_access_key, and session_token conflict with profile.")
                        sys.exit(1)



import collections
import logging
import yaml
import sys
import os
from yaml.representer import Representer

from bridgy.config.base import ConfigBase

logger = logging.getLogger()


def _readConfig():
    # ensure yaml uses a defaultdict(str)
    yaml.add_representer(collections.defaultdict,
                         Representer.represent_str)
    try:
        with open(os.path.expanduser(ConfigBase.path), 'r') as fh:
            config = yaml.load(fh)
    except Exception as ex:
        logger.error("Unable to read config (%s): %s" % (ConfigBase.path, ex))
        sys.exit(1)

    return config

def _detectConfigSchema(config):
    if 'config-schema' in config:
        return str(config['config-schema'])

    if 'inventory' in config:
        if 'source' in config['inventory']:
            source = config['inventory']['source']
            if isinstance(source, str):
                return '1'
            if isinstance(source, list):
                return '2'

    # support legacy behavior: default to original config schema if detection fails
    return '1'

def Config(initialData=None):
    if initialData != None:
        configContents = initialData
        schema = _detectConfigSchema(configContents)
    else:
        if os.path.exists(os.path.expanduser(ConfigBase.path)):
            configContents = _readConfig()
            schema = _detectConfigSchema(configContents)
        else:
            # the config does not exist and needs to be created. Use the latest schema
            schema = '2'

    if schema == '2':
        from bridgy.config.v2 import Config
        return Config(initialData)

    elif schema == '1':
        from bridgy.config.v1 import Config
        return Config(initialData)
    
    raise RuntimeError("Listed config schema is unsupported (%s)." % repr(schema))


import logging
import uuid
import sys
import os

from blessings import Terminal

term = Terminal()
logger = logging.getLogger()


class UnsupportedPlatform(Exception): pass

def platform():

    if 'linux' in sys.platform:
        return 'linux'
    elif 'darwin' in  sys.platform:
        return 'osx'
    elif 'win' in sys.platform:
        return 'windows'
    return sys.platform

class SupportedPlatforms(object):
    def __init__(self, *platforms):
        self.platforms = platforms

    def __call__(self, func):
        decorator_self = self
        def wrapper(*args, **kwargs):
            decorator_self.check_supported_platforms()
            func(*args,**kwargs)
        return wrapper

    def check_supported_platforms(self):
        normalized = platform()

        if normalized not in self.platforms:
            raise UnsupportedPlatform('Unsupported platform (%s)' % normalized)

def shortUuid():
    return str(uuid.uuid4())[:8]

def memoize(f):
    class memodict(dict):
        def __init__(self, f):
            self.f = f
        def __call__(self, *args):
            return self[args]
        def __missing__(self, key):
            ret = self[key] = self.f(*key)
            return ret
    return memodict(f)


import abc
import warnings
import collections

from bridgy.error import MissingBastionHost

with warnings.catch_warnings():
    # Thiw warns about using the slow implementation of SequenceMatcher
    # instead of the python-Levenshtein module, which requires compilation.
    # I'd prefer for users tp simply use this tool without the need to
    # compile since the search space is probably fairly small
    warnings.filterwarnings("ignore", category=UserWarning)
    from fuzzywuzzy import fuzz

Bastion = collections.namedtuple("Bastion", "destination options")
Instance = collections.namedtuple("Instance", "name address aliases source")
# allow there to be optional kwargs that default to None
Instance.__new__.__defaults__ = (None,) * len(Instance._fields)

class InventorySource(object):
    __metaclass__ = abc.ABCMeta

    name = "Invalid"
    bastion = None

    def __init__(self, *args, **kwargs):
        if 'name' in kwargs:
            self.name = "%s (%s)" % (kwargs['name'], self.name)

        if 'bastion' in kwargs:
            if 'address' not in kwargs['bastion']:
                raise MissingBastionHost

            if 'user' in kwargs['bastion']:
                destination = '{user}@{host}'.format(user=kwargs['bastion']['user'],
                                                     host=kwargs['bastion']['address'])
            else:
                destination = kwargs['bastion']['address']

            bastion_options = ''
            if 'options' in kwargs['bastion']:
                bastion_options = kwargs['bastion']['options']
            
            self.bastion = Bastion(destination=destination, options=bastion_options)

    @abc.abstractmethod
    def update(self): pass

    @abc.abstractmethod
    def instances(self, stub=True): pass

    def search(self, targets, partial=True, fuzzy=False):
        allInstances = self.instances()
        matchedInstances = set()

        for host in targets:
            for instance in allInstances:
                names = [instance.name]
                if instance.aliases != None:
                    names += list(instance.aliases)
                for name in names:
                    if host.lower() == name.lower():
                        matchedInstances.add((100, instance))
                    elif partial and host.lower() in name.lower():
                        matchedInstances.add((99, instance))
                    
                    if fuzzy:
                        score = fuzz.partial_ratio(host.lower(), name.lower())
                        if score > 85 or host.lower() in name.lower():
                            matchedInstances.add((score, instance))

        # it is possible for the same instance to be matched, if so, it should only
        # appear on the return list once (still ordered by the most probable match)
        return list(collections.OrderedDict([(v, None) for k, v in sorted(list(matchedInstances))]).keys())


class InventorySet(InventorySource):

    def __init__(self, inventories=None, **kwargs):
        super(InventorySet, self).__init__(inventories, **kwargs)
        self.inventories = []

        if inventories != None:
            if not isinstance(inventories, list) and not isinstance(inventories, tuple):
                raise RuntimeError("InventorySet only takes a list of inventories. Given: %s" % repr(type(inventories)))

            for inventory in inventories:
                self.add(inventory)

    def add(self, inventory):
        if not isinstance(inventory, InventorySource):
            raise RuntimeError("InventorySet item is not an inventory. Given: %s" % repr(type(inventory)))

        self.inventories.append(inventory)

    @property
    def name(self):
        return " + ".join([inventory.name for inventory in self.inventories])

    def update(self):
        for inventory in self.inventories:
            inventory.update()

    def instances(self, stub=True):
        instances = []

        for inventory in self.inventories:
            instances.extend(inventory.instances())

        return instances

    def search(self, targets, partial=True, fuzzy=False):
        instances = []

        for inventory in self.inventories:
            instances.extend(inventory.search(targets, partial, fuzzy))

        return instances


from yaml.representer import Representer
import collections
import logging
import yaml
import sys
import os

from bridgy import inventory
from bridgy.config.base import ConfigBase

logger = logging.getLogger()

CONFIG_TEMPLATE_NAME = 'sample_config_1.yml'


class Config(ConfigBase):

    def sources(self):
        srcCfg = self.dig('inventory', 'source')
        if srcCfg:
            return [ (srcCfg['type'], srcCfg,) ]
        return []

    @property
    def version(self):
        return 1

    @property
    def config_template_path(self):
        return CONFIG_TEMPLATE_NAME

    def verify(self):
        super(Config, self).verify()

        for source, srcCfg in self.sources():
            if source not in srcCfg.keys():
                logger.error("No inventory-specific section specified for %s source (%s):" % (repr(source), self.path))
                sys.exit(1)





import os
import sys
import logging
from bridgy.error import *
from bridgy.inventory import get_bastion
from bridgy.utils import platform, UnsupportedPlatform

logger = logging.getLogger(__name__)

def run(cmd):
    logger.debug(cmd)
    return os.system(cmd)

class Sshfs(object):

    def __init__(self, config, instance, remotedir=None, dry_run=False):
        if not hasattr(config, '__getitem__'):
            raise BadConfigError
        if not isinstance(instance, tuple):
            raise BadInstanceError

        self.config = config
        self.instance = instance
        self.remotedir = remotedir
        self.dry_run = dry_run

    @staticmethod
    def ensure_sshfs_installed():
        if run('which sshfs >/dev/null 2>&1') != 0:
            logger.error("SSHFS is not installed")
            sys.exit(1)

    @property
    def destination(self):
        if self.config.dig('ssh', 'user'):
            return '{user}@{host}'.format(user=self.config.dig('ssh', 'user'),
                                          host=self.instance.address)
        else:
            return self.instance.address

    @property
    def options(self):
        bastion = ''
        options = ''

        bastionObj = get_bastion(self.config, self.instance)

        if bastionObj != None:
            template = "-o ProxyCommand='ssh {options} -W %h:%p {destination}'"
            bastion = template.format(options=bastionObj.options,
                                      destination=bastionObj.destination)

        options = self.config.dig('sshfs', 'options') or ''

        return '{} {}'.format(bastion, options)

    @property
    def command(self):
        cmd = 'sshfs {options} {destination}:{remotedir} {mountpoint}'

        return cmd.format(destination=self.destination,
                          remotedir=self.remotedir,
                          mountpoint=self.mountpoint,
                          options=self.options)

    @classmethod
    def mounts(cls, mount_root_dir):
        _platform = platform()

        if _platform == 'osx':
            lines = [s.split() for s in os.popen("df -Ph").read().splitlines()][1:]
            system_mounts = set([fields[-1] for fields in lines if ':' in fields[0]])
        elif _platform == 'linux':
            lines = [line.strip("\n").split(" ") for line in open("/etc/mtab", "r").readlines()]
            system_mounts = set([mp for src, mp, fs, opt, p1, p2 in lines if fs == "fuse.sshfs"])
        else:
            raise UnsupportedPlatform

        possible_owned_mounts = set([os.path.join(mount_root_dir, d) for d in os.listdir(mount_root_dir)])
        owned_mounts = system_mounts & possible_owned_mounts
        return list(owned_mounts)

    @property
    def is_mounted(self):
        return self.mountpoint in Sshfs.mounts(self.config.mount_root_dir)

    @property
    def mountpoint(self):
        return os.path.join(self.config.mount_root_dir, '%s@%s' % (self.instance.name, self.instance.address))

    def mount(self):
        if not self.remotedir:
            raise BadRemoteDir("No remotedir specified")

        if not os.path.exists(self.mountpoint):
            os.mkdir(self.mountpoint)

        if self.is_mounted:
            logger.warn("Already mounted at %s" % self.mountpoint)
            sys.exit(1)

        if self.dry_run:
            logger.debug(self.command)
            return True

        rc = run(self.command)
        if rc == 0:
            return True

        logger.error("Failed to mount instance {} ({}) at {}".format(self.instance.name, self.instance.address, self.mountpoint))
        os.rmdir(self.mountpoint)
        return False

    def unmount(self, mountpoint=None):
        if not mountpoint:
            mountpoint = self.mountpoint

        if self.dry_run:
            logger.debug(self.command)
            return

        _platform = platform()
        if _platform == 'osx':
            umount_cmd = 'umount'
        elif _platform == 'linux':
            umount_cmd = 'fusermount -u'
        else:
            raise UnsupportedPlatform

        if os.path.exists(mountpoint) and run("%s %s" % (umount_cmd, mountpoint)) == 0:
            os.rmdir(mountpoint)
            return True
        return False



import os
import re
import sys
import logging
from functools import partial

from bridgy.utils import memoize
from bridgy.error import MissingBastionHost
from bridgy.inventory.source import Bastion, Instance, InventorySet
from bridgy.inventory.aws import AwsInventory
from bridgy.inventory.flatfile import CsvInventory
from bridgy.inventory.newrelic import NewRelicInventory
# from gcp import GcpInventory

logger = logging.getLogger()

SOURCES = {
    'aws': AwsInventory,
    # 'gcp': GcpInventory,
    'csv': CsvInventory,
    'newrelic': NewRelicInventory,
}


@memoize
def inventory(config):
    inventorySet = InventorySet()

    for source, srcCfg in config.sources():
        if source == 'aws':
            # the cache directory for the original v1 config did not separate 
            # out multiple aws profiles into subdirectories
            if config.version == 1:
                cache_dir = config.inventoryDir(AwsInventory.name)
            else:
                cache_dir = config.inventoryDir(AwsInventory.name, srcCfg['name'])
            
            if not os.path.exists(cache_dir):
                os.mkdir(cache_dir)

            from botocore.exceptions import ProfileNotFound
            try:
                inv = AwsInventory(cache_dir, **srcCfg)
                inventorySet.add(inv)
            except ProfileNotFound:
                logger.error("Unconfigured AWS profile configured.")
                sys.exit(1)

        elif source == 'csv':
            inv = CsvInventory(path=config.inventoryDir(source, srcCfg['file']), **srcCfg)
            inventorySet.add(inv)

        elif source == 'newrelic':

            proxies = {}

            if config.dig('inventory', 'http_proxy'):
                proxies['http'] = config.dig('inventory', 'http_proxy')
            elif 'HTTP_PROXY' in os.environ:
                proxies['http'] = os.environ['HTTP_PROXY']
            elif 'http_proxy' in os.environ:
                proxies['http'] = os.environ['http_proxy']

            if config.dig('inventory', 'https_proxy'):
                proxies['https'] = config.dig('inventory', 'https_proxy')
            elif 'HTTPS_PROXY' in os.environ:
                proxies['https'] = os.environ['HTTPS_PROXY']
            elif 'https_proxy' in os.environ:
                proxies['https'] = os.environ['https_proxy']

            inv = NewRelicInventory(data_path=config.inventoryDir(NewRelicInventory.name),
                                    proxies=proxies,
                                    **srcCfg)
 
            inventorySet.add(inv)
        
    return inventorySet

def instance_filter(instance, include_re=None, exclude_re=None):
    comparables = [instance.name, instance.address]

    if instance.aliases:
        comparables.extend(list(instance.aliases))

    if include_re:
        for name in comparables:
            if include_re.search(name):
                return True
        return False
    elif exclude_re:
        for name in comparables:
            if exclude_re.search(name):
                return False
        return True
    else:
        return True

@memoize
def instances(config):
    include_re, exclude_re = None, None
    if config.dig('inventory', 'include_pattern'):
        include_re = re.compile(config.dig('inventory', 'include_pattern'))
    if config.dig('inventory', 'exclude_pattern'):
        exclude_re = re.compile(config.dig('inventory', 'exclude_pattern'))

    all_instances = inventory(config).instances()
    config_instance_filter = partial(instance_filter, include_re=include_re, exclude_re=exclude_re)
    return list(filter(config_instance_filter, all_instances))

@memoize
def get_bastion(config, instance):
    bastion = None
    
    for inv in inventory(config).inventories:
        if inv.name == instance.source and inv.bastion != None:
            bastion = inv.bastion

    if bastion == None and 'bastion' in config:
        if not config.dig('bastion', 'address'):
            raise MissingBastionHost

        # build a destination from possible config combinations
        if config.dig('bastion', 'user'):
            destination = '{user}@{host}'.format(user=config.dig('bastion', 'user'),
                                                 host=config.dig('bastion', 'address'))
        else:
            destination = config.dig('bastion', 'address')

        bastion_options = config.dig('bastion', 'options') or ''

        bastion = Bastion(destination=destination, options=bastion_options)

    return bastion

def search(config, targets):
    fuzzy = False
    if config.dig('inventory', 'fuzzy_search'):
        fuzzy = config.dig('inventory', 'fuzzy_search')

    include_re, exclude_re = None, None
    if config.dig('inventory', 'include_pattern'):
        include_re = re.compile(config.dig('inventory', 'include_pattern'))
    if config.dig('inventory', 'exclude_pattern'):
        exclude_re = re.compile(config.dig('inventory', 'exclude_pattern'))

    matched_instances = inventory(config).search(targets, fuzzy=fuzzy)
    config_instance_filter = partial(instance_filter, include_re=include_re, exclude_re=exclude_re)
    return list(filter(config_instance_filter, matched_instances))

def update(config):
    inventory(config).update()



import os
import boto3
import placebo
import logging

from bridgy.inventory.source import InventorySource, Instance

logger = logging.getLogger()

class AwsInventory(InventorySource):

    name = 'aws'

    # kwargs: access_key_id, secret_access_key, session_token, region, profile, config_path
    def __init__(self, cache_dir, **kwargs):
        super(AwsInventory, self).__init__(cache_dir, **kwargs)

        # this is an override for the config location (at least useful for testing)
        if 'config_path' in kwargs:
            os.environ['AWS_CONFIG_FILE'] = os.path.join(kwargs['config_path'], "config")
            os.environ['AWS_SHARED_CREDENTIALS_FILE'] = os.path.join(kwargs['config_path'], "credentials")

        if 'profile' in kwargs and kwargs['profile'] != None:
            session = boto3.Session(
                profile_name=kwargs['profile'],
                region_name=kwargs['region']
            )
        elif 'access_key_id' in kwargs and 'secret_access_key' in kwargs and 'session_token' in kwargs and 'region' in kwargs:
            session = boto3.Session(
                aws_access_key_id=kwargs['access_key_id'],
                aws_secret_access_key=kwargs['secret_access_key'],
                aws_session_token=kwargs['session_token'],
                region_name=kwargs['region']
            )
        else:
            # pull from ~/.aws/* configs (or other boto search paths)
            session = boto3.Session()

        self.pill = placebo.attach(session, data_path=cache_dir)

        self.client = session.client('ec2')

    def update(self):
        try:
            self.__ec2_search(stub=False)
        except KeyboardInterrupt:
            logger.error("Cancelled by user")

    def instances(self):
        data = self.__ec2_search(stub=True)

        instances = []
        for reservation in data['Reservations']:
            for instance in reservation['Instances']:

                # try to find the best dns/ip address to reach this box
                address = None
                if instance['PublicDnsName']:
                    address = instance['PublicDnsName']
                elif instance['PrivateIpAddress']:
                    address = instance['PrivateIpAddress']

                # try to find the best field to match a name against
                aliases = list()
                if 'Tags' in list(instance.keys()):
                    for tagDict in instance['Tags']:
                        if tagDict['Key'] == 'Name':
                            aliases.append(tagDict['Value'])
                            break

                if instance['PublicDnsName']:
                    aliases.append(instance['PublicDnsName'])
                if instance['PrivateDnsName']:
                    aliases.append(instance['PrivateDnsName'])
                if instance['InstanceId']:
                    aliases.append(instance['InstanceId'])

                aliases[:] = [x for x in aliases if x != None]
                name = aliases.pop(0)

                # take note of this instance
                if name != None and address != None:
                    if len(aliases) > 0:
                        instances.append(Instance(name, address, tuple(aliases), self.name))
                    else:
                        instances.append(Instance(name, address, self.name))

        return instances

    def __ec2_search(self, tag=None, value=None, stub=True):
        filters = []
        if value:
            filters.append({
                'Name': 'tag:' + tag,
                'Values': [value]
            })

        if stub:
            self.pill.playback()
            data = self.client.describe_instances(Filters=filters)
        else:
            self.pill.record()
            data = self.client.describe_instances(Filters=filters)
            self.pill.stop()
        return data
